/*
 * Copyright (c) 2017 Louis Langholtz https://github.com/louis-langholtz/PlayRho
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

/*
 * Notes:
 *   - Any code in here must, for now, be C++14 standard compliant code.
 *   - I'm of the opinion that short of looking at the resultant assembly, it's hard
 *     to say/know/tell what the compiler actually optimzies or doesn't.
 *   - `benchmark::DoNotOptimize` seemingly only prevents enclosed expressions from
 *     being totally optimized away and has no effect on avoiding sub-expression
 *     optimization especially in regards to output from constexpr functions.
 *   - I've opted to use "random" data to help prevent optimizations that might make
 *     timing meaningless. This incurs the time overhead of generating the random value
 *     which then must be factored in to analysis of the output results.
 */

#include <benchmark/benchmark.h>

#include <PlayRho/Common/Math.hpp>

#include <PlayRho/Dynamics/World.hpp>
#include <PlayRho/Dynamics/StepConf.hpp>

#include <PlayRho/Dynamics/Contacts/ContactSolver.hpp>
#include <PlayRho/Dynamics/Contacts/VelocityConstraint.hpp>
#include <PlayRho/Dynamics/Joints/RevoluteJoint.hpp>

#include <PlayRho/Collision/Manifold.hpp>
#include <PlayRho/Collision/WorldManifold.hpp>
#include <PlayRho/Collision/ShapeSeparation.hpp>
#include <PlayRho/Collision/Shapes/PolygonShape.hpp>
#include <PlayRho/Collision/Shapes/DiskShape.hpp>

static float RandomFloat(float lo, float hi)
{
    constexpr auto RAND_LIMIT = 32767;
    auto r = static_cast<float>(std::rand() & (RAND_LIMIT));
    r /= RAND_LIMIT;
    r = (hi - lo) * r + lo;
    return r;
}

static void FloatAdd(benchmark::State& state)
{
    const auto a = 22.93f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    const auto b = 91.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(a + b);
    }
}

static void FloatMult(benchmark::State& state)
{
    const auto a = 22.93f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    const auto b = 91.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(a * b);
    }
}

static void FloatDiv(benchmark::State& state)
{
    const auto a = 5.93f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    const auto b = -4.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(a / b);
    }
}

static void FloatSqrt(benchmark::State& state)
{
    const auto a = 15.91f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(std::sqrt(a));
    }
}

static void FloatSin(benchmark::State& state)
{
    const auto a = 15.91f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(std::sin(a));
    }
}

static void FloatCos(benchmark::State& state)
{
    const auto b = -4.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(std::cos(b));
    }
}

static void FloatAtan2(benchmark::State& state)
{
    const auto a = 15.91f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    const auto b = -4.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(std::atan2(a, b));
    }
}

static void FloatSinCos(benchmark::State& state)
{
    const auto b = -4.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        // If runtime of sin + cos = sin or cos then seemingly hardware
        //   calculates both at same time and compiler knows that.
        benchmark::DoNotOptimize(std::sin(b));
        benchmark::DoNotOptimize(std::cos(b));
    }
}

static void FloatAlmostEqual1(benchmark::State& state)
{
    const auto x = static_cast<float>(rand() - (RAND_MAX / 2)) / static_cast<float>(RAND_MAX / 2);
    const auto y = static_cast<float>(rand() - (RAND_MAX / 2)) / static_cast<float>(RAND_MAX / 2);
    const auto ulp = static_cast<int>(rand() % 8);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize((playrho::Abs(x - y) < (std::numeric_limits<float>::epsilon() * playrho::Abs(x + y) * ulp)) || playrho::AlmostZero(x - y));
    }
}

static void FloatAlmostEqual2(benchmark::State& state)
{
    const auto x = static_cast<float>(rand() - (RAND_MAX / 2)) / static_cast<float>(RAND_MAX / 2);
    const auto y = static_cast<float>(rand() - (RAND_MAX / 2)) / static_cast<float>(RAND_MAX / 2);
    const auto ulp = static_cast<unsigned>(rand() % 8);
    while (state.KeepRunning())
    {
        // Accesses the floats as unsigned 32 bit ints and strips off the signbits.
        const auto nX = (*reinterpret_cast<const std::uint32_t*>(&x)) & 0x7FFFFFF;
        const auto nY = (*reinterpret_cast<const std::uint32_t*>(&y)) & 0x7FFFFFF;
        // Checks if difference between the greater 32-bit unsigned int and the lesser is
        // less than or equal to the ULP value.
        benchmark::DoNotOptimize(((nX >= nY)? nX - nY: nY - nX) <= ulp);
    }
}

static void FloatPositiveDivTrunc(benchmark::State& state)
{
    constexpr auto oneRotation = 2 * playrho::Pi;
    const auto x = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    auto turns = 0.0f;
    auto wholeTurns = 0.0f;
    auto remainder = 0.0f;
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(turns = x / oneRotation);
        benchmark::DoNotOptimize(wholeTurns = std::trunc(turns));
        benchmark::DoNotOptimize(remainder = turns - wholeTurns);
    }
}

static void FloatPositiveDivModf(benchmark::State& state)
{
    constexpr auto oneRotation = 2 * playrho::Pi;
    const auto x = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    auto turns = 0.0f;
    auto wholeTurns = 0.0f;
    auto remainder = 0.0f;
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(turns = x / oneRotation);
        benchmark::DoNotOptimize(remainder = std::modf(x, &wholeTurns));
    }
}

static void FloatPositiveFmod(benchmark::State& state)
{
    constexpr auto oneRotation = 2 * playrho::Pi;
    const auto x = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(std::fmod(x, oneRotation));
    }
}

static void AABB(benchmark::State& state)
{
    const auto pui0 = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    const auto pui1 = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    const auto pui2 = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    const auto pui3 = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    const auto p0 = playrho::Length2D{pui0 * playrho::Meter, 1.0f * playrho::Meter};
    const auto p1 = playrho::Length2D{pui1 * playrho::Meter, 2.0f * playrho::Meter};
    const auto p2 = playrho::Length2D{pui2 * playrho::Meter, 1.0f * playrho::Meter};
    const auto p3 = playrho::Length2D{pui3 * playrho::Meter, 2.0f * playrho::Meter};
    
    while (state.KeepRunning())
    {
        const auto aabb0 = playrho::AABB{p0, p1};
        const auto aabb1 = playrho::AABB{p2, p3};
        benchmark::DoNotOptimize(playrho::TestOverlap(aabb0, aabb1));
        benchmark::DoNotOptimize(playrho::Contains(aabb0, aabb1));
    }
}

// ----

static void LengthSquaredViaDotProduct(benchmark::State& state)
{
    const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto v1 = playrho::Vec2(a, b);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::Dot(v1, v1));
    }
}

static void GetLengthSquared(benchmark::State& state)
{
    const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetLengthSquared(playrho::Vec2(a, b)));
    }
}

static void GetLength(benchmark::State& state)
{
    const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetLength(playrho::Vec2(a, b)));
    }
}

static void hypot(benchmark::State& state)
{
    const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(std::hypot(a, b));
    }
}

static void UnitVectorFromVector(benchmark::State& state)
{
    const auto a = static_cast<float>(rand() - (RAND_MAX / 2)) * 22.93f / static_cast<float>(RAND_MAX);
    const auto b = static_cast<float>(rand() - (RAND_MAX / 2)) * 91.1092f / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetUnitVector(playrho::Vec2{a, b}));
    }
}

static void UnitVectorFromVectorAndBack(benchmark::State& state)
{
    const auto a = static_cast<float>(rand() - (RAND_MAX / 2)) * 22.93f / static_cast<float>(RAND_MAX);
    const auto b = static_cast<float>(rand() - (RAND_MAX / 2)) * 91.1092f / static_cast<float>(RAND_MAX);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetVec2(playrho::GetUnitVector(playrho::Vec2{a, b})));
    }
}

static void UnitVecFromAngle(benchmark::State& state)
{
    const auto a = static_cast<float>(rand()) * (23.1f * playrho::Radian * playrho::Pi / static_cast<float>(RAND_MAX));
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::UnitVec2::Get(a));
    }
}

static void DifferentSignsViaSignbit(benchmark::State& state)
{
    const auto a = static_cast<float>(rand() - (RAND_MAX / 2));
    const auto b = static_cast<float>(rand() - (RAND_MAX / 2));
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(std::signbit(a) != std::signbit(b));
    }
}

static void DifferentSignsViaMultiplication(benchmark::State& state)
{
    const auto a = static_cast<float>(rand() - (RAND_MAX / 2));
    const auto b = static_cast<float>(rand() - (RAND_MAX / 2));
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(a * b < 0.0f);
    }
}

// ----

static void FloatAddTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 22.93f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        const auto b = 91.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(a + b);
    }
}

static void FloatMultTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 22.93f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        const auto b = 91.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(a * b);
    }
}

static void FloatDivTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 5.93f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        const auto b = -4.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(a / b);
    }
}

static void FloatSqrtTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 15.91f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        const auto b = 91.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(std::sqrt(a));
        benchmark::DoNotOptimize(std::sqrt(b));
    }
}

static void FloatSinTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 15.91f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        const auto b = -4.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(std::sin(a));
        benchmark::DoNotOptimize(std::sin(b));
    }
}

static void FloatCosTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 15.91f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        const auto b = -4.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(std::cos(a));
        benchmark::DoNotOptimize(std::cos(b));
    }
}

static void FloatAtan2TwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 15.91f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        const auto b = -4.1092f * static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(std::atan2(a, b));
        benchmark::DoNotOptimize(std::atan2(b, a));
    }
}

static void FloatAlmostEqualThreeRand1(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto x = static_cast<float>(rand() - (RAND_MAX / 2)) / static_cast<float>(RAND_MAX / 2);
        const auto y = static_cast<float>(rand() - (RAND_MAX / 2)) / static_cast<float>(RAND_MAX / 2);
        const auto ulp = static_cast<int>(rand() % 8);
        benchmark::DoNotOptimize((playrho::Abs(x - y) <
                                  (std::numeric_limits<float>::epsilon() *
                                   playrho::Abs(x + y) * ulp)) || playrho::AlmostZero(x - y));
    }
}

static void DoubleAddTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 22.93 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        const auto b = 91.1092 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        benchmark::DoNotOptimize(a + b);
    }
}

static void DoubleMultTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 22.93 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        const auto b = 91.1092 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        benchmark::DoNotOptimize(a * b);
    }
}

static void DoubleDivTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 5.93 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        const auto b = -4.1092 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        benchmark::DoNotOptimize(a / b);
    }
}

static void DoubleSqrtTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 15.91 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        const auto b = 91.1092 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        benchmark::DoNotOptimize(std::sqrt(a));
        benchmark::DoNotOptimize(std::sqrt(b));
    }
}

static void DoubleSinTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 15.91 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        const auto b = -4.1092 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        benchmark::DoNotOptimize(std::sin(a));
        benchmark::DoNotOptimize(std::sin(b));
    }
}

static void DoubleCosTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 15.91 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        const auto b = -4.1092 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        benchmark::DoNotOptimize(std::cos(a));
        benchmark::DoNotOptimize(std::cos(b));
    }
}

static void DoubleAtan2TwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = 15.91 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        const auto b = -4.1092 * static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        benchmark::DoNotOptimize(std::atan2(a, b));
        benchmark::DoNotOptimize(std::atan2(b, a));
    }
}

// ---

static void LengthSquaredViaDotProductTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto v1 = playrho::Vec2(a, b);
        benchmark::DoNotOptimize(playrho::Dot(v1, v1));
    }
}

static void GetLengthSquaredTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        benchmark::DoNotOptimize(playrho::GetLengthSquared(playrho::Vec2(a, b)));
    }
}

static void GetLengthTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        benchmark::DoNotOptimize(playrho::GetLength(playrho::Vec2(a, b)));
    }
}

static void hypotTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        benchmark::DoNotOptimize(std::hypot(a, b));
    }
}

static void UnitVectorFromVectorTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = static_cast<float>(rand() - (RAND_MAX / 2)) * 22.93f / static_cast<float>(RAND_MAX);
        const auto b = static_cast<float>(rand() - (RAND_MAX / 2)) * 91.1092f / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(playrho::GetUnitVector(playrho::Vec2{a, b}));
    }
}

static void UnitVectorFromVectorAndBackTwoRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = static_cast<float>(rand() - (RAND_MAX / 2)) * 22.93f / static_cast<float>(RAND_MAX);
        const auto b = static_cast<float>(rand() - (RAND_MAX / 2)) * 91.1092f / static_cast<float>(RAND_MAX);
        benchmark::DoNotOptimize(playrho::GetVec2(playrho::GetUnitVector(playrho::Vec2{a, b})));
    }
}

// ---

static void TwoRandValues(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX));
        benchmark::DoNotOptimize(playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX));
    }
}

static void ThreeRandValues(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::Real(15.91) * static_cast<playrho::Real>(rand()) /
                                 static_cast<playrho::Real>(RAND_MAX));
        benchmark::DoNotOptimize(playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) /
                                 static_cast<playrho::Real>(RAND_MAX));
        benchmark::DoNotOptimize(static_cast<float>(rand() - (RAND_MAX / 2)) /
                                 static_cast<float>(RAND_MAX / 2));
    }
}

static void FourRandValues(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX));
        benchmark::DoNotOptimize(playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX));
        benchmark::DoNotOptimize(playrho::Real(81.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX));
        benchmark::DoNotOptimize(playrho::Real(+0.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX));
    }
}

// ----

static void DotProduct(benchmark::State& state)
{
    const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto v1 = playrho::Vec2(a, b);
    const auto c = playrho::Real(81.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto d = playrho::Real(+0.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto v2 = playrho::Vec2(c, d);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::Dot(v1, v2));
    }
}

static void CrossProduct(benchmark::State& state)
{
    const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto v1 = playrho::Vec2(a, b);
    const auto c = playrho::Real(81.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto d = playrho::Real(+0.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
    const auto v2 = playrho::Vec2(c, d);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::Cross(v1, v2));
    }
}

// ----

static void DotProductFourRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto v1 = playrho::Vec2(a, b);
        const auto c = playrho::Real(81.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto d = playrho::Real(+0.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto v2 = playrho::Vec2(c, d);
        benchmark::DoNotOptimize(playrho::Dot(v1, v2));
    }
}

static void CrossProductFourRand(benchmark::State& state)
{
    while (state.KeepRunning())
    {
        const auto a = playrho::Real(15.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto b = playrho::Real(-4.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto v1 = playrho::Vec2(a, b);
        const auto c = playrho::Real(81.91) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto d = playrho::Real(+0.1092) * static_cast<playrho::Real>(rand()) / static_cast<playrho::Real>(RAND_MAX);
        const auto v2 = playrho::Vec2(c, d);
        benchmark::DoNotOptimize(Cross(v1, v2));
    }
}

static void MaxSepBetweenRelRectanglesNoStop(benchmark::State& state)
{
    const auto rot0 = playrho::Angle{playrho::Real{45.0f} * playrho::Degree};
    const auto xfm0 = playrho::Transformation{playrho::Vec2{0, -2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::Get(rot0)}; // bottom
    const auto xfm1 = playrho::Transformation{playrho::Vec2{0, +2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::GetRight()}; // top
    
    const auto dim = playrho::Real(2) * playrho::Meter;
    const auto shape0 = playrho::PolygonShape(dim, dim);
    const auto shape1 = playrho::PolygonShape(dim, dim);
    
    // Rotate square A and put it below square B.
    // In ASCII art terms:
    //
    //   +---4---+
    //   |   |   |
    //   | B 3   |
    //   |   |   |
    //   |   2   |
    //   |   |   |
    //   |   1   |
    //   |  /+\  |
    //   2-1-*-1-2
    //    /  1  \
    //   / A |   \
    //  +    2    +
    //   \   |   /
    //    \  3  /
    //     \ | /
    //      \4/
    //       +
    
    const auto child0 = shape0.GetChild(0);
    const auto child1 = shape1.GetChild(0);
    
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child0, xfm0, child1, xfm1));
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child1, xfm1, child0, xfm0));
    }
}

static void MaxSepBetweenRelRectangles(benchmark::State& state)
{
    const auto rot0 = playrho::Angle{playrho::Real{45.0f} * playrho::Degree};
    const auto xfm0 = playrho::Transformation{playrho::Vec2{0, -2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::Get(rot0)}; // bottom
    const auto xfm1 = playrho::Transformation{playrho::Vec2{0, +2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::GetRight()}; // top

    const auto dim = playrho::Real(2) * playrho::Meter;
    const auto shape0 = playrho::PolygonShape(dim, dim);
    const auto shape1 = playrho::PolygonShape(dim, dim);
    
    // Rotate square A and put it below square B.
    // In ASCII art terms:
    //
    //   +---4---+
    //   |   |   |
    //   | B 3   |
    //   |   |   |
    //   |   2   |
    //   |   |   |
    //   |   1   |
    //   |  /+\  |
    //   2-1-*-1-2
    //    /  1  \
    //   / A |   \
    //  +    2    +
    //   \   |   /
    //    \  3  /
    //     \ | /
    //      \4/
    //       +

    const auto child0 = shape0.GetChild(0);
    const auto child1 = shape1.GetChild(0);
    const auto totalRadius = child0.GetVertexRadius() + child1.GetVertexRadius();

    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child0, xfm0, child1, xfm1, totalRadius));
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child1, xfm1, child0, xfm0, totalRadius));
    }
}

static void MaxSepBetweenRelRectangles2NoStop(benchmark::State& state)
{
    const auto dim = playrho::Real(2) * playrho::Meter;
    const auto shape = playrho::PolygonShape(dim, dim);
    
    const auto xfm0 = playrho::Transformation{
        playrho::Vec2{0, -1} * (playrho::Real(1) * playrho::Meter),
        playrho::UnitVec2::GetRight()
    }; // bottom
    const auto xfm1 = playrho::Transformation{
        playrho::Vec2{0, +1} * (playrho::Real(1) * playrho::Meter),
        playrho::UnitVec2::GetRight()
    }; // top
    
    const auto child0 = shape.GetChild(0);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child0, xfm0, child0, xfm1));
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child0, xfm1, child0, xfm0));
    }
}

static void MaxSepBetweenRelRectangles2(benchmark::State& state)
{
    const auto dim = playrho::Real(2) * playrho::Meter;
    const auto shape = playrho::PolygonShape(dim, dim);
    
    const auto xfm0 = playrho::Transformation{
        playrho::Vec2{0, -1} * (playrho::Real(1) * playrho::Meter),
        playrho::UnitVec2::GetRight()
    }; // bottom
    const auto xfm1 = playrho::Transformation{
        playrho::Vec2{0, +1} * (playrho::Real(1) * playrho::Meter),
        playrho::UnitVec2::GetRight()
    }; // top
    
    const auto child0 = shape.GetChild(0);
    const auto totalRadius = child0.GetVertexRadius() * playrho::Real(2);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child0, xfm0, child0, xfm1, totalRadius));
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child0, xfm1, child0, xfm0, totalRadius));
    }
}

static void MaxSepBetweenRel4x4(benchmark::State& state)
{
    const auto rot0 = playrho::Angle{playrho::Real{45.0f} * playrho::Degree};
    const auto xfm0 = playrho::Transformation{playrho::Vec2{0, -2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::Get(rot0)}; // bottom
    const auto xfm1 = playrho::Transformation{playrho::Vec2{0, +2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::GetRight()}; // top
    
    const auto dim = playrho::Real(2) * playrho::Meter;
    const auto shape0 = playrho::PolygonShape(dim, dim);
    const auto shape1 = playrho::PolygonShape(dim, dim);
    
    // Rotate square A and put it below square B.
    // In ASCII art terms:
    //
    //   +---4---+
    //   |   |   |
    //   | B 3   |
    //   |   |   |
    //   |   2   |
    //   |   |   |
    //   |   1   |
    //   |  /+\  |
    //   2-1-*-1-2
    //    /  1  \
    //   / A |   \
    //  +    2    +
    //   \   |   /
    //    \  3  /
    //     \ | /
    //      \4/
    //       +
    
    const auto child0 = shape0.GetChild(0);
    const auto child1 = shape1.GetChild(0);

    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetMaxSeparation4x4(child0, xfm0, child1, xfm1));
        benchmark::DoNotOptimize(playrho::GetMaxSeparation4x4(child1, xfm1, child0, xfm0));
    }
}

static void MaxSepBetweenRel2_4x4(benchmark::State& state)
{
    const auto dim = playrho::Real(2) * playrho::Meter;
    const auto shape = playrho::PolygonShape(dim, dim);
    
    const auto xfm0 = playrho::Transformation{
        playrho::Vec2{0, -1} * (playrho::Real(1) * playrho::Meter),
        playrho::UnitVec2::GetRight()
    }; // bottom
    const auto xfm1 = playrho::Transformation{
        playrho::Vec2{0, +1} * (playrho::Real(1) * playrho::Meter),
        playrho::UnitVec2::GetRight()
    }; // top
    
    const auto child0 = shape.GetChild(0);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetMaxSeparation4x4(child0, xfm0, child0, xfm1));
        benchmark::DoNotOptimize(playrho::GetMaxSeparation4x4(child0, xfm1, child0, xfm0));
    }
}

static void MaxSepBetweenAbsRectangles(benchmark::State& state)
{
    const auto rot0 = playrho::Angle{playrho::Real{45.0f} * playrho::Degree};
    const auto xfm0 = playrho::Transformation{playrho::Vec2{0, -2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::Get(rot0)}; // bottom
    const auto xfm1 = playrho::Transformation{playrho::Vec2{0, +2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::GetRight()}; // top

    const auto dim = playrho::Real(2) * playrho::Meter;
    const auto shape0 = playrho::PolygonShape{dim, dim}.Transform(xfm0);
    const auto shape1 = playrho::PolygonShape{dim, dim}.Transform(xfm1);
    
    // Rotate square A and put it below square B.
    // In ASCII art terms:
    //
    //   +---4---+
    //   |   |   |
    //   | B 3   |
    //   |   |   |
    //   |   2   |
    //   |   |   |
    //   |   1   |
    //   |  /+\  |
    //   2-1-*-1-2
    //    /  1  \
    //   / A |   \
    //  +    2    +
    //   \   |   /
    //    \  3  /
    //     \ | /
    //      \4/
    //       +
    
    const auto child0 = shape0.GetChild(0);
    const auto child1 = shape1.GetChild(0);
    const auto totalRadius = child0.GetVertexRadius() + child1.GetVertexRadius();
    
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child0, child1, totalRadius));
        benchmark::DoNotOptimize(playrho::GetMaxSeparation(child1, child0, totalRadius));
    }
}

static void ManifoldForTwoSquares1(benchmark::State& state)
{
    const auto dim = playrho::Real(2) * playrho::Meter;
    
    // creates a square
    const auto shape = playrho::PolygonShape(dim, dim);
    
    const auto rot0 = playrho::Angle{playrho::Real{45.0f} * playrho::Degree};
    const auto xfm0 = playrho::Transformation{playrho::Vec2{0, -2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::Get(rot0)}; // bottom
    const auto xfm1 = playrho::Transformation{playrho::Vec2{0, +2} * (playrho::Real(1) * playrho::Meter), playrho::UnitVec2::GetRight()}; // top
    
    // Rotate square A and put it below square B.
    // In ASCII art terms:
    //
    //   +---4---+
    //   |   |   |
    //   | B 3   |
    //   |   |   |
    //   |   2   |
    //   |   |   |
    //   |   1   |
    //   |  /+\  |
    //   2-1-*-1-2
    //    /  1  \
    //   / A |   \
    //  +    2    +
    //   \   |   /
    //    \  3  /
    //     \ | /
    //      \4/
    //       +

    while (state.KeepRunning())
    {
        // CollideShapes(shape.GetChild(0), xfm0, shape.GetChild(0), xfm1);
        benchmark::DoNotOptimize(playrho::CollideShapes(shape.GetChild(0), xfm0, shape.GetChild(0), xfm1));
    }
}

static void ManifoldForTwoSquares2(benchmark::State& state)
{
    // Shape A: square
    const auto shape0 = playrho::PolygonShape(playrho::Real{2} * playrho::Meter, playrho::Real{2} * playrho::Meter);
    
    // Shape B: wide rectangle
    const auto shape1 = playrho::PolygonShape(playrho::Real{3} * playrho::Meter, playrho::Real{1.5f} * playrho::Meter);
    
    const auto xfm0 = playrho::Transformation{
        playrho::Vec2{-2, 0} * (playrho::Real(1) * playrho::Meter),
        playrho::UnitVec2::GetRight()
    }; // left
    const auto xfm1 = playrho::Transformation{
        playrho::Vec2{+2, 0} * (playrho::Real(1) * playrho::Meter),
        playrho::UnitVec2::GetRight()
    }; // right
    
    // Put square left, wide rectangle right.
    // In ASCII art terms:
    //
    //   +-------2
    //   |     +-+---------+
    //   |   A | 1   B     |
    //   |     | |         |
    //   4-3-2-1-*-1-2-3-4-5
    //   |     | |         |
    //   |     | 1         |
    //   |     +-+---------+
    //   +-------2
    //
    while (state.KeepRunning())
    {
        //CollideShapes(shape0.GetChild(0), xfm0, shape1.GetChild(0), xfm1);
        benchmark::DoNotOptimize(CollideShapes(shape0.GetChild(0), xfm0, shape1.GetChild(0), xfm1));
    }
}

static void ConstructAndAssignVC(benchmark::State& state)
{
    const auto friction = playrho::Real(0.5);
    const auto restitution = playrho::Real(1);
    const auto tangentSpeed = playrho::LinearVelocity{playrho::Real(1.5) * playrho::MeterPerSecond};
    const auto invMass = playrho::Real(1) / playrho::Kilogram;
    const auto invRotI = playrho::Real(1) / ((playrho::SquareMeter * playrho::Kilogram) / playrho::SquareRadian);
    const auto normal = playrho::UnitVec2::GetRight();
    const auto location = playrho::Length2D{playrho::Real(0) * playrho::Meter, playrho::Real(0) * playrho::Meter};
    const auto impulse = playrho::Momentum2D{playrho::Momentum{0}, playrho::Momentum{0}};
    const auto separation = playrho::Length{playrho::Real(-0.001) * playrho::Meter};
    const auto ps0 = playrho::WorldManifold::PointData{location, impulse, separation};
    const auto worldManifold = playrho::WorldManifold{normal, ps0};
    
    const auto locA = playrho::Length2D{playrho::Real(+1) * playrho::Meter, playrho::Real(0) * playrho::Meter};
    const auto posA = playrho::Position{locA, playrho::Angle(0)};
    const auto velA = playrho::Velocity{
        playrho::LinearVelocity2D{playrho::Real(-0.5) * playrho::MeterPerSecond, playrho::Real(0) * playrho::MeterPerSecond},
        playrho::AngularVelocity{playrho::Real(0) * playrho::RadianPerSecond}
    };
    
    const auto locB = playrho::Length2D{playrho::Real(-1) * playrho::Meter, playrho::Real(0) * playrho::Meter};
    const auto posB = playrho::Position{locB, playrho::Angle(0)};
    const auto velB = playrho::Velocity{
        playrho::LinearVelocity2D{playrho::Real(+0.5) * playrho::MeterPerSecond, playrho::Real(0) * playrho::MeterPerSecond},
        playrho::AngularVelocity{playrho::Real(0) * playrho::RadianPerSecond}
    };
    
    auto bcA = playrho::BodyConstraint{invMass, invRotI, locA, posA, velA};
    auto bcB = playrho::BodyConstraint{invMass, invRotI, locB, posB, velB};
    
    auto vc = playrho::VelocityConstraint{};
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(vc = playrho::VelocityConstraint{friction, restitution, tangentSpeed, worldManifold, bcA, bcB});
    }
}

static void malloc_free_random_size(benchmark::State& state)
{
    auto sizes = std::array<size_t, 100>();
    for (auto& size: sizes)
    {
        size = (static_cast<std::size_t>(std::rand()) % std::size_t{1ul << 18ul}) + std::size_t{1ul};
    }

    auto i = std::size_t{0};
    auto p = static_cast<void*>(nullptr);
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(p = std::malloc(sizes[i]));
        std::free(p);
        i = (i < (sizes.max_size() - 1ul))? i + 1: 0;
    }
}

static void random_malloc_free_100(benchmark::State& state)
{
    auto sizes = std::array<size_t, 100>();
    for (auto& size: sizes)
    {
        size = (static_cast<std::size_t>(std::rand()) % std::size_t{1ul << 18ul}) + std::size_t{1ul};
    }
    auto pointers = std::array<void*, 100>();
    while (state.KeepRunning())
    {
        auto ptr = std::begin(pointers);
        for (auto size: sizes)
        {
            benchmark::DoNotOptimize(*ptr = std::malloc(size));
            ++ptr;
        }
        for (auto& p: pointers)
        {
            std::free(p);
        }
    }
}

static void SolveVC(benchmark::State& state)
{
    const auto friction = playrho::Real(0.5);
    const auto restitution = playrho::Real(1);
    const auto tangentSpeed = playrho::LinearVelocity{playrho::Real(1.5) * playrho::MeterPerSecond};
    const auto invMass = playrho::Real(1) / playrho::Kilogram;
    const auto invRotI = playrho::Real(1) / ((playrho::SquareMeter * playrho::Kilogram) / playrho::SquareRadian);
    const auto normal = playrho::UnitVec2::GetRight();
    const auto location = playrho::Length2D{playrho::Real(0) * playrho::Meter, playrho::Real(0) * playrho::Meter};
    const auto impulse = playrho::Momentum2D{playrho::Momentum{0}, playrho::Momentum{0}};
    const auto separation = playrho::Length{playrho::Real(-0.001) * playrho::Meter};
    const auto ps0 = playrho::WorldManifold::PointData{location, impulse, separation};
    const auto worldManifold = playrho::WorldManifold{normal, ps0};
    
    const auto locA = playrho::Length2D{playrho::Real(+1) * playrho::Meter, playrho::Real(0) * playrho::Meter};
    const auto posA = playrho::Position{locA, playrho::Angle(0)};
    const auto velA = playrho::Velocity{
        playrho::LinearVelocity2D{playrho::Real(-0.5) * playrho::MeterPerSecond, playrho::Real(0) * playrho::MeterPerSecond},
        playrho::AngularVelocity{playrho::Real(0) * playrho::RadianPerSecond}
    };

    const auto locB = playrho::Length2D{playrho::Real(-1) * playrho::Meter, playrho::Real(0) * playrho::Meter};
    const auto posB = playrho::Position{locB, playrho::Angle(0)};
    const auto velB = playrho::Velocity{
        playrho::LinearVelocity2D{playrho::Real(+0.5) * playrho::MeterPerSecond, playrho::Real(0) * playrho::MeterPerSecond},
        playrho::AngularVelocity{playrho::Real(0) * playrho::RadianPerSecond}
    };

    auto bcA = playrho::BodyConstraint{invMass, invRotI, locA, posA, velA};
    auto bcB = playrho::BodyConstraint{invMass, invRotI, locB, posB, velB};

    auto vc = playrho::VelocityConstraint{friction, restitution, tangentSpeed, worldManifold, bcA, bcB};
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(playrho::GaussSeidel::SolveVelocityConstraint(vc));
        benchmark::ClobberMemory();
    }
}

static void DefaultWorldStep(benchmark::State& state)
{
    auto world = playrho::World{};
    const auto stepConf = playrho::StepConf{};
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(world.Step(stepConf));
    }
}

static void DropDisks(benchmark::State& state)
{
    auto world = playrho::World{playrho::WorldDef{}.UseGravity(playrho::EarthlyGravity)};

    const auto diskDef = playrho::DiskShape::Conf{}.UseVertexRadius(0.5f * playrho::Meter);
    const auto shape = std::make_shared<playrho::DiskShape>(diskDef);
    for (auto i = 0; i < state.range(); ++i)
    {
        const auto x = i * 2 * playrho::Meter;
        const auto location = playrho::Length2D{x, 0 * playrho::Meter};
        const auto body = world.CreateBody(playrho::BodyDef{}
                                           .UseType(playrho::BodyType::Dynamic)
                                           .UseLocation(location));
        body->CreateFixture(shape);
    }

    const auto stepConf = playrho::StepConf{};
    while (state.KeepRunning())
    {
        benchmark::DoNotOptimize(world.Step(stepConf));
    }
}

static void AddPairStressTest(benchmark::State& state, int count)
{
    const auto diskConf = playrho::DiskShape::Conf{}
        .UseVertexRadius(playrho::Meter / 10)
        .UseDensity(0.01f * playrho::KilogramPerSquareMeter);
    const auto diskShape = std::make_shared<playrho::DiskShape>(diskConf);

    const auto polygonConf = playrho::PolygonShape::Conf{}.UseDensity(1.0f * playrho::KilogramPerSquareMeter);
    const auto rectShape = std::make_shared<playrho::PolygonShape>(1.5f * playrho::Meter, 1.5f * playrho::Meter, polygonConf);
    
    const auto rectBodyDef = playrho::BodyDef{}
        .UseType(playrho::BodyType::Dynamic)
        .UseBullet(true)
        .UseLocation(playrho::Length2D{-40.0f * playrho::Meter, 5.0f * playrho::Meter})
        .UseLinearVelocity(playrho::LinearVelocity2D{playrho::Vec2(150.0f, 0.0f) * playrho::MeterPerSecond});

    const auto worldDef = playrho::WorldDef{}.UseGravity(playrho::LinearAcceleration2D{}).UseInitialTreeSize(8192);
    const auto stepConf = playrho::StepConf{};
    const auto minX = -6.0f;
    const auto maxX = 0.0f;
    const auto minY = 4.0f;
    const auto maxY = 6.0f;
    const auto bd = playrho::BodyDef{}.UseType(playrho::BodyType::Dynamic);
    while (state.KeepRunning())
    {
        state.PauseTiming();
        auto world = playrho::World{worldDef};
        {
            for (auto i = 0; i < count; ++i)
            {
                const auto location = playrho::Vec2(RandomFloat(minX, maxX), RandomFloat(minY, maxY)) * playrho::Meter;
                const auto body = world.CreateBody(playrho::BodyDef{bd}.UseLocation(location));
                body->CreateFixture(diskShape);
            }
        }
        world.CreateBody(rectBodyDef)->CreateFixture(rectShape);
        for (auto i = 0; i < state.range(); ++i)
        {
            world.Step(stepConf);
        }
        state.ResumeTiming();
        
        world.Step(stepConf);
    }
}

static void AddPairStressTest400(benchmark::State& state)
{
    AddPairStressTest(state, 400);
}

static void DropTiles(int count)
{
    const auto linearSlop = playrho::Meter / 1000;
    const auto angularSlop = (playrho::Pi * 2 * playrho::Radian) / 180;
    const auto vertexRadius = linearSlop * 2;
    const auto conf = playrho::PolygonShape::Conf{}.UseVertexRadius(vertexRadius);
    auto m_world = playrho::World{
        playrho::WorldDef{}.UseMinVertexRadius(vertexRadius).UseInitialTreeSize(8192)
    };
    
    {
        const auto a = playrho::Real{0.5f};
        const auto ground = m_world.CreateBody(playrho::BodyDef{}.UseLocation(playrho::Length2D{0, -a * playrho::Meter}));
        
        const auto N = 200;
        const auto M = 10;
        playrho::Length2D position;
        GetY(position) = playrho::Real(0.0f) * playrho::Meter;
        for (auto j = 0; j < M; ++j)
        {
            GetX(position) = -N * a * playrho::Meter;
            for (auto i = 0; i < N; ++i)
            {
                auto shape = playrho::PolygonShape{conf};
                SetAsBox(shape, a * playrho::Meter, a * playrho::Meter, position, playrho::Angle{0});
                ground->CreateFixture(std::make_shared<playrho::PolygonShape>(shape));
                GetX(position) += 2.0f * a * playrho::Meter;
            }
            GetY(position) -= 2.0f * a * playrho::Meter;
        }
    }
    
    {
        const auto a = playrho::Real{0.5f};
        const auto shape = std::make_shared<playrho::PolygonShape>(a * playrho::Meter, a * playrho::Meter, conf);
        shape->SetDensity(playrho::Real{5} * playrho::KilogramPerSquareMeter);
        
        playrho::Length2D x(playrho::Real(-7.0f) * playrho::Meter, playrho::Real(0.75f) * playrho::Meter);
        playrho::Length2D y;
        const auto deltaX = playrho::Length2D(playrho::Real(0.5625f) * playrho::Meter, playrho::Real(1.25f) * playrho::Meter);
        const auto deltaY = playrho::Length2D(playrho::Real(1.125f) * playrho::Meter, playrho::Real(0.0f) * playrho::Meter);
        
        for (auto i = 0; i < count; ++i)
        {
            y = x;
            
            for (auto j = i; j < count; ++j)
            {
                const auto body = m_world.CreateBody(playrho::BodyDef{}.UseType(playrho::BodyType::Dynamic).UseLocation(y));
                body->CreateFixture(shape);
                y += deltaY;
            }
            
            x += deltaX;
        }
    }
    
    auto step = playrho::StepConf{};
    step.SetTime(playrho::Time{playrho::Second / playrho::Real{60}});
    step.linearSlop = linearSlop;
    step.regMinSeparation = -linearSlop * playrho::Real(3);
    step.toiMinSeparation = -linearSlop * playrho::Real(1.5f);
    step.targetDepth = linearSlop * playrho::Real(3);
    step.tolerance = linearSlop / playrho::Real(4);
    step.maxLinearCorrection = linearSlop * playrho::Real(40);
    step.maxAngularCorrection = angularSlop * playrho::Real{4};
    step.aabbExtension = linearSlop * playrho::Real(20);
    step.maxTranslation = playrho::Length{playrho::Meter * playrho::Real(4)};
    step.velocityThreshold = (playrho::Real{8} / playrho::Real{10}) * playrho::MeterPerSecond;
    step.maxSubSteps = std::uint8_t{48};

    while (GetAwakeCount(m_world) > 0)
    {
        m_world.Step(step);
    }
}

static void TilesComesToRest(benchmark::State& state)
{
    const auto range = state.range();
    while (state.KeepRunning())
    {
        DropTiles(range);
    }
}

class Tumbler
{
public:
    Tumbler();
    void Step();
    void AddSquare();
    bool IsWithin(const playrho::AABB& aabb) const;

private:
    static playrho::Body* CreateEnclosure(playrho::World& world);
    static playrho::RevoluteJoint* CreateRevoluteJoint(playrho::World& world,
                                                       playrho::Body* stable, playrho::Body* turn);

    playrho::World m_world;
    playrho::StepConf m_stepConf;
    playrho::Length m_squareLen = 0.125f * playrho::Meter;
    std::shared_ptr<playrho::PolygonShape> m_square =
        std::make_shared<playrho::PolygonShape>(m_squareLen, m_squareLen);
};

Tumbler::Tumbler()
{
    const auto g = m_world.CreateBody(playrho::BodyDef{}.UseType(playrho::BodyType::Static));
    const auto b = CreateEnclosure(m_world);
    CreateRevoluteJoint(m_world, g, b);
}

playrho::Body* Tumbler::CreateEnclosure(playrho::World& world)
{
    const auto b = world.CreateBody(playrho::BodyDef{}.UseType(playrho::BodyType::Dynamic)
                                    .UseLocation(playrho::Vec2(0, 10) * playrho::Meter)
                                    .UseAllowSleep(false));
    
    playrho::PolygonShape shape;
    shape.SetDensity(5 * playrho::KilogramPerSquareMeter);
    playrho::SetAsBox(shape, 0.5f * playrho::Meter, 10.0f * playrho::Meter, playrho::Vec2( 10.0f, 0.0f) * playrho::Meter, playrho::Angle{0});
    b->CreateFixture(std::make_shared<playrho::PolygonShape>(shape));
    playrho::SetAsBox(shape, 0.5f * playrho::Meter, 10.0f * playrho::Meter, playrho::Vec2(-10.0f, 0.0f) * playrho::Meter, playrho::Angle{0});
    b->CreateFixture(std::make_shared<playrho::PolygonShape>(shape));
    playrho::SetAsBox(shape, 10.0f * playrho::Meter, 0.5f * playrho::Meter, playrho::Vec2(0.0f, 10.0f) * playrho::Meter, playrho::Angle{0});
    b->CreateFixture(std::make_shared<playrho::PolygonShape>(shape));
    playrho::SetAsBox(shape, 10.0f * playrho::Meter, 0.5f * playrho::Meter, playrho::Vec2(0.0f, -10.0f) * playrho::Meter, playrho::Angle{0});
    b->CreateFixture(std::make_shared<playrho::PolygonShape>(shape));
    
    return b;
}

playrho::RevoluteJoint* Tumbler::CreateRevoluteJoint(playrho::World& world,
                                                     playrho::Body* stable, playrho::Body* turn)
{
    playrho::RevoluteJointDef jd;
    jd.bodyA = stable;
    jd.bodyB = turn;
    jd.localAnchorA = playrho::Vec2(0.0f, 10.0f) * playrho::Meter;
    jd.localAnchorB = playrho::Vec2(0.0f, 0.0f) * playrho::Meter;
    jd.referenceAngle = playrho::Angle{0};
    
    // Make it turn 4 times faster than Testbed Tumbler demo
    jd.motorSpeed = 0.2f * playrho::Pi * playrho::RadianPerSecond;

    jd.maxMotorTorque = playrho::Real{100000} * playrho::NewtonMeter; // 1e8f;
    jd.enableMotor = true;
    return static_cast<playrho::RevoluteJoint*>(world.CreateJoint(jd));
}

void Tumbler::Step()
{
    m_world.Step(m_stepConf);
}

void Tumbler::AddSquare()
{
    const auto b = m_world.CreateBody(playrho::BodyDef{}
                                      .UseType(playrho::BodyType::Dynamic)
                                      .UseLocation(playrho::Vec2(0, 10) * playrho::Meter));
    b->CreateFixture(m_square);
}

bool Tumbler::IsWithin(const playrho::AABB& aabb) const
{
    return playrho::Contains(aabb, GetAABB(m_world.GetTree()));
}

static void TumblerAddSquaresForSteps(benchmark::State& state,
                                      int squareAddingSteps, int additionalSteps)
{
    const auto rangeX = playrho::Interval<playrho::Length>{
        -15 * playrho::Meter, +15 * playrho::Meter
    };
    const auto rangeY = playrho::Interval<playrho::Length>{
        -5 * playrho::Meter, +25 * playrho::Meter
    };
    const auto aabb = playrho::AABB{rangeX, rangeY};
    while (state.KeepRunning())
    {
        Tumbler tumbler;
        for (auto i = 0; i < squareAddingSteps; ++i)
        {
            tumbler.Step();
            tumbler.AddSquare();
        }
        for (auto i = 0; i < additionalSteps; ++i)
        {
            tumbler.Step();
        }
        if (!tumbler.IsWithin(aabb))
        {
            std::cout << "escaped!" << std::endl;
            continue;
        }
    }
}

static void TumblerAdd100SquaresPlus100Steps(benchmark::State& state)
{
    TumblerAddSquaresForSteps(state, 100, 100);
}

static void TumblerAdd200SquaresPlus200Steps(benchmark::State& state)
{
    TumblerAddSquaresForSteps(state, 200, 200);
}

#if 0
#define ADD_BM(n, f) \
    BENCHMARK_PRIVATE_DECLARE(f) = \
        benchmark::RegisterBenchmark(n, f);
#endif

BENCHMARK(FloatAdd);
BENCHMARK(FloatMult);
BENCHMARK(FloatDiv);

BENCHMARK(FloatAlmostEqual1);
BENCHMARK(FloatAlmostEqual2);

BENCHMARK(DifferentSignsViaSignbit);
BENCHMARK(DifferentSignsViaMultiplication);

BENCHMARK(FloatPositiveDivTrunc);
BENCHMARK(FloatPositiveDivModf);
BENCHMARK(FloatPositiveFmod);
BENCHMARK(FloatSqrt);
BENCHMARK(FloatSin);
BENCHMARK(FloatCos);
BENCHMARK(FloatSinCos);
BENCHMARK(FloatAtan2);

BENCHMARK(AABB);

BENCHMARK(DotProduct);
BENCHMARK(CrossProduct);
BENCHMARK(LengthSquaredViaDotProduct);
BENCHMARK(GetLengthSquared);
BENCHMARK(GetLength);
BENCHMARK(hypot);
BENCHMARK(UnitVectorFromVector);
BENCHMARK(UnitVectorFromVectorAndBack);
BENCHMARK(UnitVecFromAngle);

BENCHMARK(TwoRandValues);

BENCHMARK(FloatAddTwoRand);
BENCHMARK(FloatMultTwoRand);
BENCHMARK(FloatDivTwoRand);
BENCHMARK(FloatSqrtTwoRand);
BENCHMARK(FloatSinTwoRand);
BENCHMARK(FloatCosTwoRand);
BENCHMARK(FloatAtan2TwoRand);

BENCHMARK(ThreeRandValues);
BENCHMARK(FloatAlmostEqualThreeRand1);

BENCHMARK(DoubleAddTwoRand);
BENCHMARK(DoubleMultTwoRand);
BENCHMARK(DoubleDivTwoRand);
BENCHMARK(DoubleSqrtTwoRand);
BENCHMARK(DoubleSinTwoRand);
BENCHMARK(DoubleCosTwoRand);
BENCHMARK(DoubleAtan2TwoRand);

BENCHMARK(LengthSquaredViaDotProductTwoRand);
BENCHMARK(GetLengthSquaredTwoRand);
BENCHMARK(GetLengthTwoRand);
BENCHMARK(hypotTwoRand);
BENCHMARK(UnitVectorFromVectorTwoRand);
BENCHMARK(UnitVectorFromVectorAndBackTwoRand);

BENCHMARK(FourRandValues);

BENCHMARK(DotProductFourRand);
BENCHMARK(CrossProductFourRand);

BENCHMARK(ConstructAndAssignVC);
BENCHMARK(SolveVC);
BENCHMARK(DefaultWorldStep);

BENCHMARK(MaxSepBetweenAbsRectangles);
BENCHMARK(MaxSepBetweenRel4x4);
BENCHMARK(MaxSepBetweenRel2_4x4);
BENCHMARK(MaxSepBetweenRelRectanglesNoStop);
BENCHMARK(MaxSepBetweenRelRectangles2NoStop);
BENCHMARK(MaxSepBetweenRelRectangles);
BENCHMARK(MaxSepBetweenRelRectangles2);

BENCHMARK(ManifoldForTwoSquares1);
BENCHMARK(ManifoldForTwoSquares2);

BENCHMARK(malloc_free_random_size);

BENCHMARK(DropDisks)->Arg(0)->Arg(1)->Arg(2)->Arg(10)->Arg(100)->Arg(1000);

BENCHMARK(random_malloc_free_100);

BENCHMARK(TumblerAdd100SquaresPlus100Steps);
BENCHMARK(TumblerAdd200SquaresPlus200Steps);
BENCHMARK(AddPairStressTest400)->Arg(0)->Arg(10)->Arg(15)->Arg(16)->Arg(17)->Arg(18)->Arg(19)->Arg(20)->Arg(30);

BENCHMARK(TilesComesToRest)->Arg(12)->Arg(20)->Arg(36);

BENCHMARK_MAIN()
